"""
The :py:mod:`pce` module contains classes to acquire the uncertainty
characterization of the stochastic parameters, to generate the random
experiment and to construct the Polynomial Chaos Expansion.
"""

import os
import multiprocessing as mp
from functools import partial
import itertools
import numpy as np
from scipy import stats, special
import sobol
import warnings
import pandas as pd

class Data:
    """

    The class includes methods to create a file to store the samples
    and to store information on the stochastic design space, exctracted
    from the :file:`design_space.txt` and :file:`stochastic_space.txt` files.

    Parameters
    ----------
    inputs : dict
        input dictionary with information on
        the uncertainty quantification.
    space_obj : object
        object that contains information on the test case.

    """

    def __init__(self, inputs, space_obj):
        self.inputs = inputs
        self.path = os.path.split(
            os.path.dirname(
                os.path.abspath(__file__)))[0]
        self.space_obj = space_obj
        self.path_res = None

        self.stoch_data = {}
        self.filename_samples = ''

    def create_samples_file(self):
        """
        Creating the file that saved the input samples and model outputs.
        """

        # results directory
        self.path_res = os.path.join(self.path,
                                     'RESULTS',
                                     self.space_obj.case,
                                     'UQ',
                                     self.inputs['results dir'],
                                     )

        # if results directory does not exist, create one
        if not os.path.exists(self.path_res):
            os.makedirs(self.path_res)

        # file where evaluated samples are stored
        self.filename_samples = os.path.join(self.path_res, 'samples.csv')
        
        if not os.path.isfile(self.filename_samples):
            df = pd.DataFrame([self.stoch_data['names'] + 
                           self.inputs['objective names']])
            with open(self.filename_samples, 'w') as f:
                 df.to_csv(f, header=False, index=False, line_terminator='\n')
        
        '''
        if not os.path.isfile(self.filename_samples):
            with open(self.filename_samples, "w") as file:
                self.stoch_data['names'] + \
                        self.inputs['objective names']:
                    file.write('%25s, ' % name)
                file.write('\n')
        '''
        
    def read_stoch_parameters(self, var_values=[]):
        """
        Read in the stochastic design space
        and save the information in a dictionary

        Parameters
        ----------
        var_values : list, optional
            The design variable values in case of robust optimization.
            The default is [].

        """

        # check that no design variables are defined in the design space,
        # for which no corresponding deterministic values are provided in
        # var_values (only when performing robust optimization)
        if len(var_values) != len(self.space_obj.var_dict):
            raise ValueError(
                """When performing UQ, make sure that no design variableS
                   are present in design_space. Each variable should be
                   converted into a model parameter.""")

        # add values in var_values list to the variable names defined in the
        # var dict (only when performing robust optimization)
        tmp = {}
        for i, key in enumerate(self.space_obj.var_dict):
            tmp[key] = var_values[i]

        # dictionary with the deterministic parameters and deterministic values
        # for the design variables
        det_dict = {**self.space_obj.par_dict, **tmp}

        # create lists to store the bounds, mean and deviation for each
        # uncertain parameter
        l_b, u_b, mean, deviation = [], [], [], []
        for key in self.space_obj.upar_dict:
            # based on relation between deviation and mean, generate values
            # for the bounds
            if self.space_obj.upar_dict[key][0] == 'absolute':
                l_b.append(det_dict[key] - self.space_obj.upar_dict[key][2])
                u_b.append(det_dict[key] + self.space_obj.upar_dict[key][2])
                deviation.append(self.space_obj.upar_dict[key][2])
            elif self.space_obj.upar_dict[key][0] == 'relative':
                l_b.append(det_dict[key] *
                           (1. - self.space_obj.upar_dict[key][2]))
                u_b.append(det_dict[key] *
                           (1. + self.space_obj.upar_dict[key][2]))
                deviation.append(self.space_obj.upar_dict[key][2] *
                                 det_dict[key])
            else:
                raise ValueError(""" The relation of the deviation to the mean
                                     should be 'relative' or 'absolute'.""")

            mean.append(det_dict[key])

        # the dictionary with information on the uncertain parameters
        self.stoch_data = {'names': list(self.space_obj.upar_dict.keys()),
                           'types': [x[1] for x in list(
                                     self.space_obj.upar_dict.values())],
                           'l_b': l_b,
                           'u_b': u_b,
                           'mean': mean,
                           'deviation': deviation}


class RandomExperiment(Data):
    """
    RandomExperiment objects include information on the random samples,
    such as dimension, scaled and unscaled values. In addition, the
    samples can be evaluated and the deterministic model outputs are
    stored.

    Parameters
    ----------
    my_data : object
        Data object with information on the uncertain parameters.
    objective_position : int
        Index of the quantity of interest in the list of available
        model outputs.

    """

    def __init__(self, my_data, objective_position):
        self.my_data = my_data
        self.objective_position = objective_position

        self.dimension = len(my_data.stoch_data['types'])
        self.dists = [None] * len(my_data.stoch_data['types'])
        self.polytypes = [None] * len(my_data.stoch_data['types'])
        self.size = None
        self.x_u = None
        self.x_u_scaled = None

        self.y = None

        self.x_prev = None
        self.y_prev = None
        self.n_samples = None

    def n_terms(self):
        '''
        This method sets the number of samples to 2*(p+n)!/p!n!,
        i.e the number of terms in the full PC
        expansion of order p in n random variables.

        '''
        # number of uncertain parameters
        n_par = len(self.my_data.stoch_data['mean'])

        # polynomial order
        p_order = self.my_data.inputs['pol order']
        result = 1.
        mmin = min(n_par, p_order)
        for i in range(mmin):
            result *= p_order + n_par - i
        result_terms = int(result / np.math.factorial(mmin))

        # number of samples for training the PCE
        self.n_samples = 2 * result_terms

    def read_previous_samples(self, create_only_samples):
        """
        Read the previously evaluated samples
        and store them for future PCE construction.

        Parameters
        ----------
        create_only_samples : bool
            the boolean that indicates if samples
            should only be created.

        """
        
        df = pd.read_csv(self.my_data.filename_samples)
        df_x = df[self.my_data.stoch_data['names']]
        df_y = df[self.my_data.inputs['objective of interest']]
        
        self.x_prev = df_x.to_numpy()
        
        self.y_prev = np.zeros((len(self.x_prev), 1))
        y_int = df_y.to_numpy()
        
        for i,y in enumerate(y_int):
            self.y_prev[i] = y
            
        '''
        with open(self.my_data.filename_samples, 'r') as file:
            lines = file.readlines()
            x_int = np.zeros((len(lines) - 1,
                              len(self.my_data.stoch_data['mean'])))
            y_int = np.zeros((len(lines) - 1, 1))

            # if samples are present in the samples file
            if len(lines) > 1:
                if (len(lines[-1].split()) !=
                    len(self.my_data.stoch_data['mean']) +
                        len(self.my_data.inputs['objective names'])):
                    raise SyntaxError(
                        """The samples file is not properly formatted
                           or it already contains samples
                           without model output.""")

                if create_only_samples:
                    raise ValueError(
                        """The samples file already contains samples
                           with model output.
                           Consider changing the result directory
                           or switching "create only samples" to False.""")

                # read in samples and deterministic model output
                for i, line in enumerate(lines[1:]):
                    line_split = line.split()
                    x_int[i] = [float(el) for el in line_split[:len(
                        self.my_data.stoch_data['mean'])]]
                    y_int[i] = float(
                        line_split[len(self.my_data.stoch_data['mean']) +
                                   self.objective_position])

        # store samples and deterministic output
        self.y_prev = np.array(y_int)
        self.x_prev = np.array(x_int)
        '''

    def create_distributions(self):
        """
        Create the distributions, polynomial distributions and polynomial types
        based on the stochastic design space. The available distributions
        are Uniform and Gaussian distributions.

        """

        polydists = [None] * len(self.my_data.stoch_data['types'])

        # iterate through the uncertain parameters defined in stochastic space
        for i, j in enumerate(self.my_data.stoch_data['types']):
            if j == 'Uniform':
                self.dists[i] = stats.uniform(
                    self.my_data.stoch_data['mean'][i] -
                    self.my_data.stoch_data['deviation'][i],
                    2. * self.my_data.stoch_data['deviation'][i])

                # scaled distribution for polynomial
                polydists[i] = stats.uniform(-1., 2.)

                # Legendre polynomial for Uniform distributions
                self.polytypes[i] = 'Legendre'

            elif j == 'Gaussian':
                self.dists[i] = stats.norm(
                    self.my_data.stoch_data['mean'][i],
                    self.my_data.stoch_data['deviation'][i])

                # scaled distribution for polynomial
                polydists[i] = stats.norm(0., 1.)

                # Legendre polynomial for Uniform distributions
                self.polytypes[i] = 'Hermite'

            else:
                raise ValueError(""" Distribution for %s not found.
                                     Choose between Uniform and Gaussian."""
                                 % self.my_data.stoch_data['types'][i])

    def create_samples(self, size=0):
        """
        Generate the samples for model evaluations. The sampling
        methods available are random sampling and Sobol sampling.
        The number of samples generated is equal to the integer
        `size`. If `size` is equal or lower than zero, no new samples
        are generated. Instead, the required samples are extracted
        from the existing samples file.

        Parameters
        ----------
        size : int, optional
            The number of newly created samples. The default is 0.

        """
        l_b = self.my_data.stoch_data['l_b']
        u_b = self.my_data.stoch_data['u_b']
        method = self.my_data.inputs['sampling method']

        # when samples need to be created
        if size > 0:
            self.x_u = np.zeros((size, self.dimension))
            if method == 'RANDOM':
                # random generation
                for i in range(self.dimension):
                    self.x_u[:, i] = self.dists[i].rvs(size)

            elif method == 'SOBOL':
                # sobol sequence
                skip = 123454
                x_tr = sobol.sample(dimension=self.dimension,
                                    n_points=size + len(self.x_prev),
                                    skip=skip)

                for i in range(self.dimension):
                    self.x_u[:, i] = self.dists[i].ppf(
                        x_tr[len(self.x_prev):, i])

            if len(self.x_prev) > 0:
                # when new samples are combined with previous samples
                self.x_u = np.concatenate((self.x_prev, self.x_u))

            self.size = (len(self.x_u))

        elif size == 0:
            # when no new samples are needed, get the ones from samples file
            self.x_u = self.x_prev
            self.size = len(self.x_u)

        else:
            # less samples needed than available in samples file
            self.x_u = np.split(self.x_prev,
                                [len(self.x_prev) + size,
                                 len(self.x_prev)])[0]
            self.size = len(self.x_u)

        # generate the scaled samples
        self.x_u_scaled = np.zeros((self.size, self.dimension))
        for i in range(self.dimension):
            for j in range(self.size):
                self.x_u_scaled[j, i] = ((self.x_u[j, i] - l_b[i]) /
                                         (u_b[i] - l_b[i]) * 2. - 1.)

    def create_only_samples(self, create_only_samples):
        """

        Add the generated samples to the samples file.
        These samples can be used for model evaluation,
        when the model is not connected to the UQ algorithm.

        Parameters
        ----------
        create_only_samples : bool
            the boolean that indicates
            if samples should only be created

        """

        if create_only_samples:

            df = pd.DataFrame(self.x_u, columns=None)
            with open(self.my_data.filename_samples, 'a+') as f:
                df.to_csv(f, header=False, index=False, line_terminator='\n')

            '''
            # add the samples to the samples file, no output results
            with open(self.my_data.filename_samples, 'a+') as file:
                for i, x_u_sample in enumerate(self.x_u):
                    line = list(x_u_sample)
                    for j in line:
                        file.write('%25f ' % j)

                    file.write('\n')
            '''

    def evaluate(self, eval_func, params):
        """
        Evaluate the samples in the model
        and store the samples and outputs in the samples file.

        """

        # number of samples needed
        size = self.n_samples - len(self.x_prev)

        # create Uniform/gaussian distributions and corresponding orthogonal
        # polynomials
        self.create_samples(size=size)

        samples = self.x_u[-size:]
        unc_samples = np.tile(
            list(self.my_data.space_obj.par_dict.values()), (size, 1))

        # list with names from variables and parameters
        par_var_list = list(self.my_data.space_obj.par_dict.keys())
        par_var_list += list(self.my_data.space_obj.var_dict.keys())
        for j, elem in enumerate(self.my_data.space_obj.upar_dict.keys()):
            # get name from uncertain parameters out of upar_dict, and get the
            # index for each name out of the complete par_var_list

            unc_idx = par_var_list.index(elem)

            # based on the retrieved index, replace the deterministic value
            # of the stochastic parameter with the random sample
            unc_samples[:, unc_idx] = samples[:, j]

        eval_dict = []
        for sample in unc_samples:

            # bring parameter and variable names and values for each sample
            # into a dictionary
            eval_dict.append(
                self.my_data.space_obj.convert_into_dictionary(sample))

        if self.my_data.inputs['n jobs'] == 1:
            # linear processing
            res = []
            for index, sample in enumerate(eval_dict):
                res.append(eval_func((index + len(self.x_prev), sample), params=params))

            df = pd.DataFrame(samples, columns=None)
            df2 = pd.DataFrame(res, columns=None)
            df3 = pd.concat([df, df2], axis=1)
                        
            with open(self.my_data.filename_samples, 'a+') as f:
                df3.to_csv(f, header=False, index=False, line_terminator='\n')

                '''
                with open(self.my_data.filename_samples, 'a+') as file:
                    line = list(samples[index]) + list(res[-1])
                    for j in line:
                        file.write('%25f ' % j)
                    file.write('\n')
                '''

        else:
            # multiprocessing
            pool = mp.Pool(processes=self.my_data.inputs['n jobs'])
            res = pool.map(
                partial(
                    eval_func,
                    params=params),
                enumerate(eval_dict))
            pool.close()

            # append new samples and model outputs to samples file
            with open(self.my_data.filename_samples, 'a+') as file:
                for i, sample in enumerate(samples):
                    line = list(np.concatenate((sample, res[i])))
                    for j in line:
                        file.write('%25f ' % j)

                    file.write('\n')

        # check that the quantity of interest exists
        if self.objective_position > len(res[0]) - 1:
            raise IndexError(""" The objective "%s" falls out of
                                 the range of predefined quantities
                                 of interest. Only %i outputs are
                                 returned from the model""" %
                             (self.my_data.inputs['objective names'][
                                 int(self.objective_position)], len(res[0])))

        # combine model output for quantity of interest with previous results
        # available in samples file
        y_res = [row[self.objective_position] for row in res]
        if self.y_prev.size:
            self.y = np.vstack((self.y_prev, np.array(y_res).reshape((-1, 1))))
        else:
            self.y = np.array(y_res).reshape((-1, 1))

class PCE(RandomExperiment):
    """

    Class which creates a Polynomial Chaos Expansion (PCE) object. A PCE is
    characterized by the following attributes:

        - Basis : PC basis functions
        - Coefficients : PC coefficients
        - Moments : Statistical moments
        - sensitivity : Sobol' indices


    Parameters
    ----------
    RandomExperiment : obj
        RandomExperiment object, with information on the random samples.

    """

    def __init__(self, Experiment):
        self.my_experiment = Experiment
        self.order = self.my_experiment.my_data.inputs['pol order']
        self.a_matrix = 0.0
        self.loo = 0.0
        self.basis = dict()
        self.coefficients = dict()
        self.sensitivity = dict()
        self.moments = dict()
        self.psi_sq = 0.0

    #####################################
    # module uncertainty quantification #
    #####################################

    def n_to_sum(self, n, s):
        '''
        This function creates a list
        of all possible vectors of length 'n' that sum to 's'

        Parameters
        ----------
        n : int
            vector's length (dimension)
        s : int
            sum (total order)

        Returns
        -------
        The output is a generator object to be called as follows
            - list(n_to_sum(dimension,i))
        '''

        if n == 1:
            yield (s,)
        else:
            for i in range(s + 1):
                for j in self.n_to_sum(n - 1, s - i):
                    yield j + (i,)

    def multindices(self, idx):
        """

        This method returns a set of multi-indices

        Parameters
        ----------
        idx : list
            the range for the number of terms in the PCE

        Returns
        -------
        list
            the list with multi-indices

        """
        dimension = self.my_experiment.dimension

        multindices = list()

        # generate the set of multi-indices
        for i in range(self.order + 1):
            multindices.extend(list(self.n_to_sum(dimension, i)))

        return [multindices[i] for i in idx]

    def ols(self, a_matrix, b_matrix):
        """

        Perform Ordinary Least Squares on the input matrices a_matrix and
        b_matrix.
        Return the result.

        Parameters
        ----------
        a_matrix : array
            the matrix with the response of the basis functions to the samples
        b_matrix : array
            the model output for the input samples

        Returns
        -------
        2_D array
            result of the ordinary least-square regression

        """

        m, n = a_matrix.shape
        deg_of_freedom = m - n

        sol = np.linalg.lstsq(a_matrix, b_matrix, rcond=None)
        sol_2 = np.column_stack(1.0 / deg_of_freedom * sol[1])

        var = np.dot(np.row_stack(np.diag(np.linalg.inv(
            np.dot(np.transpose(a_matrix), a_matrix)))), sol_2)

        return np.column_stack((sol[0], var))

    def calc_a(self, multindices):
        """

        This method builds the matrix containing the basis functions evaluated
        at sample locations, i.e. the matrix A in Au = b

        Parameters
        ----------
        multindices : list
            the list with the multi-indices

        Returns
        -------
        A : array
            the matrix with the evaluated basis functions for the samples

        """

        dimension = self.my_experiment.dimension
        x_u_scaled = self.my_experiment.x_u_scaled
        a_matrix = np.ones([self.my_experiment.size, len(multindices)])

        # generate the A matrix
        for i, multiindex in enumerate(multindices):
            for j in range(dimension):
                deg = multiindex[j]
                if self.my_experiment.polytypes[j] == 'Legendre':
                    a_matrix[:, i] *= special.eval_legendre(deg,
                                                            x_u_scaled[:, j])

                elif self.my_experiment.polytypes[j] == 'Hermite':
                    a_matrix[:, i] *= special.eval_hermitenorm(deg,
                                                               x_u_scaled[:,
                                                                          j])

        return a_matrix

    def run(self):
        """

        Solve Ordinary Least Squares problem
        Full PC expansion is assumed containing n_terms(dimension,order)


        """

        # number of terms in the PCE
        n_terms = int(self.my_experiment.n_samples / 2)

        self.basis['model'] = range(n_terms)

        # full PCE basis
        self.basis['multi-indices'] = self.multindices(self.basis['model'])
        self.basis['polytypes'] = self.my_experiment.polytypes

        # determine the A matrix
        self.a_matrix = self.calc_a(self.basis['multi-indices'])

        # get the coefficient through OLS
        self.coefficients = self.ols(self.a_matrix, self.my_experiment.y)

        # acquire the statistics
        self.get_statistics(mean=True, variance=True)

    ################################
    # module statistics extraction #
    ################################

    def get_statistics(self, mean=True, variance=True):
        '''
        This function calculates high order moments (up to order 2) by
        taking advantage of the fact that any permutation of indices
        will lead to the same value for the summand.

        Returns Statistics: mean, std. deviation

        '''

        coeff = self.coefficients

        # quantify the mean
        if mean:
            if self.basis['model'][0] == 0:
                self.moments['mean'] = coeff[0, 0]

        if variance:
            # the term <psi_i, psi_j>
            self.psi_sq = self.get_psi_sq()

            # quantify the variance
            var = 0.0
            if self.basis['model'][0] == 0:
                for i in range(1, len(self.basis['model'])):
                    var += self.psi_sq[i] * coeff[i, 0]**2

            self.moments['variance'] = var

    def get_psi_sq(self):
        """

        Calculate the term <psii,psij>

        Returns
        -------
        psi_sq : array
            the term <psii,psij>

        """
        dim = self.my_experiment.dimension

        multindices = self.basis['multi-indices']

        n_terms = len(multindices)

        psi_sq = np.ones(n_terms,)

        for i in range(n_terms):
            for j in range(dim):
                deg = multindices[i][j]

                if self.my_experiment.polytypes[j] == 'Legendre':

                    x_i, w_i = special.p_roots(deg + 1)

                    '''
                    Integrate exactly the SQUARE
                    of the Legendre polynomial. For example,
                    if the Legendre polynomial is of order (deg),
                    the numerical integration must be exact
                    till order (deg**2). Thus, we need at least
                    (deg+1) abscissas' and weights.
                    '''
                    poly = special.legendre(deg)**2
                    psi_sq[i] *= 1.0 / 2 * sum(w_i * poly(x_i))

                elif self.my_experiment.polytypes[j] == 'Hermite':

                    x_i, w_i = special.he_roots(deg + 1)

                    '''
                    special.he_roots(deg) and
                    np.polynomial.hermite_e.hermegauss(deg)
                    returns the same abscissas'
                    but different weights (!). There is a factor 2
                    between the two. Given the fact that the integral of
                    the standard Gaussian must be 1,
                    np.polynomial.hermite_e.hermegauss(deg)
                    provides the right weights.
                    '''
                    poly = special.hermitenorm(deg)**2

                    # 2*w_i*poly(x_i)
                    psi_sq[i] *= 1.0 / np.sqrt(2 * np.pi) * sum(w_i *
                                                                poly(x_i))

        return psi_sq

    ######################
    # module sensitivity #
    ######################

    def calc_sobol(self):
        '''
        This method calculates the Sobol' indices Si of PCE.
        They represent the fraction of the total variance
        that can be attributed to each input variable i (Si)
        or combinations of variables (i1,i2,...,is).
        The total Sobol indices quantify the total effect
        of variable i, i.e. including the first order effect
        and the interactions with the other variables.

        '''

        dimension = self.my_experiment.dimension

        var = self.moments['variance']

        coeff = self.coefficients

        psi_sq = self.get_psi_sq()  # vector of size (P+1)

        # Create list of all of terms
        terms = list()
        for i in range(1, self.order + 1):
            terms.extend(list(itertools.combinations(range(dimension), i)))

        copy = self.basis['multi-indices'][:]
        del copy[0]

        # Compute the Sobol' indices
        s_i = np.zeros(len(terms),)
        for i, multindex in enumerate(copy):
            for j, term in enumerate(terms):
                if (len(np.nonzero(multindex)[0]) == len(term)) and (
                        np.nonzero(multindex)[0] == term).all():
                    s_i[j] += 1.0 / var * coeff[i + 1][0]**2 * psi_sq[i + 1]

                    break

        # Compute the total Sobol' indices
        s_tot_i = np.zeros(dimension,)
        for i in range(dimension):
            idx = np.array([item for item in range(len(terms))
                            if (terms[item] == np.array([i])).any()])
            for index, j in enumerate(idx):
                s_tot_i[i] += s_i[j]

        self.sensitivity['s_i'] = s_i
        self.sensitivity['s_tot_i'] = s_tot_i
        self.sensitivity['terms'] = terms

    ##############################
    # accuracy evaluation module #
    ##############################

    def calc_loo(self):
        '''
        This method evaluates the Leave-One-Out (LOO) error for the
        constructed PCE. The formula is adopted from Sudret et al. [1].

        [1] Sudret, B. (2014). Polynomial chaos expansions and stochastic
        finite-element methods. In Risk and Reliability in Geotechnical
        Engineering.

        '''

        size = self.my_experiment.size
        y_res = self.my_experiment.y
        y_hat = np.row_stack((np.dot(self.a_matrix, self.coefficients[:, 0])))

        # acquire the diagonal matrix D
        b_matrix = np.linalg.inv(np.dot(np.transpose(self.a_matrix),
                                        self.a_matrix))
        c_matrix = np.dot(b_matrix, np.transpose(self.a_matrix))
        h_matrix = np.dot(self.a_matrix, c_matrix)
        d_matrix = np.diag(h_matrix)

        # quantify the LOO error
        self.loo = 0.
        for i in range(size):
            deltai = (y_res[i, 0] - y_hat[i, 0]) / (1 - d_matrix[i])

            self.loo += 1.0 / float(size) * deltai**2. / (np.var(y_res))

        if self.loo > 1.:
            warnings.warn("The LOO error is higher than 1. "
                          "Check the UQ characterization and results.")

    ##########################
    # result printing module #
    ##########################

    def print_res(self):
        """
        This method prints an overview of the inputs and results of the PCE.
        Additionally, it writes the PCE information and Sobol' indices
        in the result files.'
        """

        mean = self.moments['mean']
        var = self.moments['variance']

        print('%s Polynomial chaos output %s \n' % ('-' * 20, '-' * 20))
        print(' Number of input variables ::'.ljust(30) + '%d' %
              len(self.my_experiment.dists))
        print(' Maximal degree ::'.ljust(30) + '%d' %
              max([sum(i) for i in self.basis['multi-indices']]))
        print(' Size of full basis ::'.ljust(30) +
              '%d' % self.my_experiment.n_samples)
        print(' Size of sparse basis ::'.ljust(30) +
              '%d' % len(self.basis['multi-indices']) + '\n')

        print(' Full model evaluation ::'.ljust(30) +
              '%d' % self.my_experiment.size)
        print(' Leave-one-out error ::'.ljust(30) + '%s' % self.loo + '\n')

        print(' Mean value ::'.ljust(30) + '%s' % mean)
        print(' Std. deviation ::'.ljust(30) + '%s' % np.sqrt(var))
        print(' First-order Sobol indices ::'.ljust(30) +
              '%s' % np.around(self.sensitivity['s_i'], decimals=4) + '\n')
        print(' Total Sobol indices ::'.ljust(30) +
              '%s' % np.around(self.sensitivity['s_tot_i'], decimals=4) + '\n')

        print('-' * 65)

        filename_res = (
            "full_pce_order_%d_%s.txt" %
            (self.order,
             self.my_experiment.my_data.inputs['objective of interest']))

        # write statistics in the result file
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               filename_res), "w") as file:
            file.write('%25s %25f \n' % ('LOO', self.loo))
            file.write(
                '%25s %25f \n' %
                ('sparse basis', len(
                    self.basis['multi-indices'])))
            file.write('%25s %25f \n' % ('mean', mean))
            file.write(
                '%25s %25f \n' % ('std. dev.', np.sqrt(var)))
                
        total = np.array([self.my_experiment.my_data.stoch_data['names'],self.sensitivity['s_i'][:len(self.sensitivity['s_tot_i'])],self.sensitivity['s_tot_i']]).transpose()
        
        df1 = pd.DataFrame(total, columns=['name',
                 'First-order Sobol indices',
                 'Total-order Sobol indices'])

        # write sobol indices in the corresponding result file
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               filename_res[:-4] + '_Sobol_indices.csv'), "w") as file:                               
            df1.to_csv(file, index=False, line_terminator='\n')
                               
            '''
            file.write(
                '%30s %30s %30s  \n' %
                ('name',
                 'First-order Sobol indices',
                 'Total-order Sobol indices'))
            indices = np.argsort(self.sensitivity['s_tot_i'])[::-1]
            for i in indices:
                file.write(
                    '%30s %30f %30f  \n' %
                    (self.my_experiment.my_data.stoch_data['names'][i],
                     self.sensitivity['s_i'][i],
                     self.sensitivity['s_tot_i'][i]))
            '''

    def draw(self, size):
        """
        This module creates the probability density function
        and cumulative distribution function and writes
        the corresponding values to construct these values
        in the result files.

        Parameters
        ----------
        size : int
            number of samples to be created

        """

        # create samples to evaluate on the PCE
        self.my_experiment.create_samples(size=size)

        # generate A matrix and the coefficients u
        a_matrix = self.calc_a(self.basis['multi-indices'])
        x_u_scaled = np.row_stack((self.coefficients[:, 0]))

        data = np.dot(a_matrix, x_u_scaled)

        # generate the pdf
        density, bins = np.histogram(data, bins=100, density=1)
        centers = [(a + b) / 2 for a, b in zip(bins[::1], bins[1::1])]
        
        total = np.array([centers,density]).transpose()
        
        df1 = pd.DataFrame(total, columns=[self.my_experiment.my_data.inputs['objective of interest'],'probability density'])
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               ("data_pdf_%s.csv"
                                % (self.my_experiment.my_data.inputs[
                                    'objective of interest']))), "w") as f:
            df1.to_csv(f, index=False, line_terminator='\n')
        
        '''
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               ("data_pdf_%s.csv"
                                % (self.my_experiment.my_data.inputs[
                                    'objective of interest']))), "w") as file:


            file.write(
                '%25s %25s \n' %
                (self.my_experiment.my_data.inputs['objective of interest'],
                 'probability density'))
            for i, j in enumerate(centers):
                file.write('%25f %25f' % (j, density[i]))
                file.write('\n')
        '''
        
        # generate the cdf
        cdf = np.cumsum(density * np.diff(bins))

        total = np.array([centers,cdf]).transpose()
        
        df1 = pd.DataFrame(total, columns=[self.my_experiment.my_data.inputs['objective of interest'],'cumulative probability'])
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               ("data_cdf_%s.csv"
                                % (self.my_experiment.my_data.inputs[
                                    'objective of interest']))), "w") as f:
            df1.to_csv(f, index=False, line_terminator='\n')

        '''
        with open(os.path.join(self.my_experiment.my_data.path_res,
                               ("data_cdf_%s"
                                % (self.my_experiment.my_data.inputs[
                                    'objective of interest']))), "w") as file:

            file.write(
                '%25s %25s \n' %
                (self.my_experiment.my_data.inputs['objective of interest'],
                 'cumulative probability'))
            for i, j in enumerate(centers):
                file.write('%25f %25f' % (j, cdf[i]))
                file.write('\n')
        '''


'''
Module to test the hydrogen-based energy system models.
'''

import os
import pytest
import rheia
import rheia.CASES.H2_FUEL.h2_fuel as lb
import rheia.CASES.H2_POWER.h2_power as lb2
import rheia.CASES.H2_MOBILITY.h2_mobility as lb3


def test_h2_fuel():
    """
    Test the LCOH and m_h2 for a generic design in the photovoltaic-hydrogen
    system model.
    """

    path = os.path.dirname(rheia.__file__)

    # the climate file considered
    filename_climate = os.path.join(path,
                                    'CASES',
                                    'DATA',
                                    'climate',
                                    'climate_Brussels.csv')

    # the object to read in the data
    my_data = lb.ReadData(filename_climate)

    # get the solar irradiance and ambient temperature
    sol_irr, t_amb = my_data.load_climate()

    # retrieve the deterministic values for the model parameters
    parameters = my_data.load_parameters()

    # define the design to be tested
    inputs = {'n_dcdc_pv': 5.,
              'n_pemel': 4.}

    # instantiate from the Evaluation class
    my_evaluation = lb.Evaluation(sol_irr, t_amb, {**parameters, **inputs})

    # evaluate the system
    my_evaluation.evaluation()

    assert round(my_evaluation.res['lcoh'], 3) == 10.172
    assert round(my_evaluation.res['m_h2'], 3) == 115.441


def test_h2_power():
    """
    Test the LCOE and SSR for a generic design in the power-to-power
    system model.
    """

    path = os.path.dirname(rheia.__file__)

    # the climate file considered
    filename_climate = os.path.join(path,
                                    'CASES',
                                    'DATA',
                                    'climate',
                                    'climate_Brussels.csv')

    filename_demand = os.path.join(path,
                                   'CASES',
                                   'DATA',
                                   'demand',
                                   'load_Brussels_dwelling.csv')

    # the object to read in the data
    my_data = lb2.ReadData(filename_climate, filename_demand)

    # get the solar irradiance and ambient temperature
    sol_irr, t_amb = my_data.load_climate()

    # get the electric load
    load_elec = my_data.load_demand()

    # retrieve the deterministic values for the model parameters
    parameters = my_data.load_parameters()

    # define the design to be tested
    inputs = {'n_pv': 10.,
              'n_pemel': 2.,
              'n_pemfc': 1.,
              'n_tank': 100.,
              }

    # instantiate from the Evaluation class
    my_evaluation = lb2.Evaluation(
        sol_irr, t_amb, load_elec, {
            **parameters, **inputs})

    # evaluate the system
    my_evaluation.evaluation()

    assert round(my_evaluation.res['lcoe'], 3) == 484.926
    assert round(my_evaluation.res['ssr'], 3) == 0.773


def test_h2_mobility():
    """
    Test the LCOD and CI for a generic design in the power-to-mobility
    system model.
    """

    path = os.path.dirname(rheia.__file__)

    # the climate file considered
    filename_climate = os.path.join(path,
                                    'CASES',
                                    'DATA',
                                    'climate',
                                    'climate_Brussels.csv')

    # the object to read in the data
    my_data = lb3.ReadData(filename_climate)

    # get the solar irradiance and ambient temperature
    sol_irr, t_amb = my_data.load_climate()

    # retrieve the deterministic values for the model parameters
    parameters = my_data.load_parameters()

    # define the design to be tested
    inputs = {'n_pv': 1000.,
              'n_pemel': 2000.,
              'n_tank': 10000.,
              'n_h2_bus': 1.,
              }

    # instantiate from the Evaluation class
    my_evaluation = lb3.Evaluation(sol_irr, t_amb, {**parameters, **inputs})

    # evaluate the system
    my_evaluation.evaluation()

    assert round(my_evaluation.res['lcom'], 3) == 1.357
    assert round(my_evaluation.res['ci'], 3) == 1.318



'''
Module to test the det_stoch_des_space module.
'''

import pytest
from rheia.CASES.determine_stoch_des_space import StochasticDesignSpace


@pytest.fixture
def run_dict():
    """
    Generate a generic input dictionary for optimization.

    Returns
    -------
    dict_opt : dict
        Input dictionary for optimization.

    """

    dict_opt = {'case': 'H2_FUEL',
                'objectives': {'DET': (-1, 1)},
                'stop': ('BUDGET', 2000),
                'n jobs': 1,
                'population size': 20,
                'results dir': 'run_tutorial',
                }

    return dict_opt


@pytest.fixture
def input_case(run_dict):
    """
    Instantiate a StochasticDesignSpace object.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.

    Returns
    -------
    test_class : object
        StochasticDesignSpace object.

    """

    design_space = 'design_space.csv'
    opt_type = list(run_dict['objectives'].keys())[0]
    case = run_dict['case']
    test_class = StochasticDesignSpace(opt_type, case, design_space)

    return test_class


def test_case_name(input_case):
    """
    Assert the case name for the StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.case == 'H2_FUEL'


def test_design_space_name(input_case):
    """
    Assert the design_space name for the StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.design_space == 'design_space.csv'


def test_opt_type(input_case):
    """
    Assert the allocation of the optimization type for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.opt_type == 'DET'


def test_n_dim(input_case):
    """
    Assert the optimization dimension for the StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.n_dim == 2


def test_n_par(input_case):
    """
    Assert the number of deterministic parameters for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.n_par == 13


def test_par_dict(input_case):
    """
    Assert the dictionary for deterministic parameters for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.par_dict == {'n_pv': 5.,
                                   'u_sol_irr': 1,
                                   'u_t_amb': 0,
                                   'capex_pv': 475,
                                   'opex_pv': 17.5,
                                   'capex_pemel': 1750,
                                   'opex_pemel': 0.04,
                                   'life_pemel': 80000,
                                   'repl_pemel': 0.175,
                                   'capex_dcdc': 150,
                                   'opex_dcdc': 0.03,
                                   'int_rate': 0.06,
                                   'infl_rate': 0.02}


def test_var_dict(input_case):
    """
    Assert the dictionary for design variables for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """
    assert input_case.var_dict == {'n_dcdc_pv': [0., 7.],
                                   'n_pemel': [0., 7.]}


def test_lb(input_case):
    """
    Assert the design variables lower bounds for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.l_b == [0., 0.]


def test_ub(input_case):
    """
    Assert the design variables upper bounds for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.u_b == [7., 7.]


def test_upar_dict(input_case):
    """
    Assert the dictionary for stochastic model parameters for the
    StochasticDesignSpace object.

    Parameters
    ----------
    input_case : object
        StochasticDesignSpace object.

    """

    assert input_case.upar_dict == {}


def test_attach_objectives(run_dict, input_case):
    """
    Assert if objectives are attached for StochasticDesignSpace object

    Parameters
    ----------
    run_dict : dict
        Input dictionary for optimization.
    input_case : object
        StochasticDesignSpace object.

    """

    input_case.attach_objectives(run_dict['objectives']['DET'])
    assert input_case.obj == (-1, 1)



"""
The :py:mod:`determine_stoch_des_space` module contains a class to read the
design_space and stochastic_space files. In addition, functions are present
to load the case object and to evaluate the characterization of the input
dictionary.
"""

import os
import sys
import collections
import warnings

def custom_formatwarning(msg, *args, **kwargs):
    """
    Customize the format of the warning message:
    Show only the message.

    """
    # ignore everything except the message
    return str(msg) + '\n'



def check_dictionary(run_dict, uq_bool=False):
    """

    This function evaluates if the items in the input dictionary are
    properly characterized.

    Parameters
    ----------
    run_dict : dict
        The input dictionary.
    uq_bool : bool, optional
        Boolean that mentions if uncertainty quantification is considered.
        The default is False.

    """

    warnings.formatwarning = custom_formatwarning
    
    rob = False

    if not isinstance(run_dict, collections.Mapping):
        raise TypeError('The input dictionary should be a dictionary.')

    requirements = ['case',
                    'n jobs',
                    'results dir',
                    ]

    if 'n jobs' not in run_dict:
        run_dict['n jobs'] = 1

    for key in requirements:
        try:
            run_dict[key]
        except BaseException:
            raise KeyError(
                '"%s" is missing in the input dictionary.' %
                key)

    if not isinstance(run_dict['n jobs'], int):
        raise TypeError(
            'The value of the key "n jobs" should be a positive integer.')

    if not isinstance(run_dict['results dir'], str):
        raise TypeError(
            'The value of the key "results dir" should be a string.')

    if not uq_bool:
        requirements += ['objectives',
                         'stop',
                         'population size',
                         'x0',
                         'cx prob',
                         'mut prob',
                         'eta',
                         ]

        if 'x0' not in run_dict:
            run_dict['x0'] = ('AUTO', 'LHS')
            warnings.warn("If no previous results existed, the starting "
                          "population is generated using Latin Hypercube sampling.")
        if 'cx prob' not in run_dict:
            run_dict['cx prob'] = 0.9
            warnings.warn("The crossover probability is defined automatically "
                          "at 0.9")
        if 'mut prob' not in run_dict:
            run_dict['mut prob'] = 0.1
            warnings.warn("The mutation probability is defined automatically "
                          "at 0.1")
        if 'eta' not in run_dict:
            run_dict['eta'] = 0.2
            warnings.warn("The eta parameter is defined automatically at 0.2")

        for key in requirements[3:]:
            try:
                run_dict[key]
            except BaseException:
                raise KeyError(
                    '"%s" is missing in the input dictionary.' %
                    key)

        if not isinstance(run_dict['objectives'], collections.Mapping):
            raise TypeError(
                'The value of the key "objectives" should be a dictionary.')

        if not list(
                run_dict['objectives'].keys())[0] == 'DET' and not list(
                run_dict['objectives'].keys())[0] == 'ROB':
            raise ValueError(
                """Please select "DET" or "ROB" as a key
                for the "objectives" value.""")
        elif not isinstance(list(run_dict['objectives'].values())[0], tuple):
            raise TypeError(
                """The value of the key "DET" or "ROB" should be a tuple
                with the weights for the objectives
                (1 for maximization, -1 for minimization).""")
        elif (not all(isinstance(weight, int) for weight in
                      list(run_dict['objectives'].values())[0])):
            raise TypeError('The weights should be equal to 1 or -1.')
        elif (not all(abs(weight) == 1 for weight in
                      list(run_dict['objectives'].values())[0])):
            raise ValueError('The weights should be equal to 1 or -1.')

        if not isinstance(run_dict['stop'], tuple):
            raise TypeError(
                """The value of the key "stop" should be a tuple
                   with two elements.""")

        if not run_dict['stop'][0] == 'BUDGET':
            raise ValueError(
                """The first element in the tuple related to the key "stop"
                   should be equal to "BUDGET".""")

        if not isinstance(run_dict['stop'][1], int):
            raise TypeError(
                """The second element in the tuple related to the key "stop"
                   should be a positive integer.""")

        if not isinstance(run_dict['population size'], int):
            raise TypeError(
                """The value of the key "population number"
                   should be a positive integer.""")

        if not isinstance(run_dict['x0'], tuple):
            raise TypeError(
                """The value of the key "x0" should be a tuple
                   with two elements.""")

        if (not run_dict['x0'][0] == 'AUTO' and not
                run_dict['x0'][0] == 'CUSTOM'):
            raise ValueError(
                """The first element in the tuple related to the key "x0"
                   should be equal to "AUTO" or "CUSTOM".""")
        elif (run_dict['x0'][0] == 'AUTO' and not
              run_dict['x0'][1] == 'RANDOM' and not
              run_dict['x0'][1] == 'LHS'):
            raise ValueError(
                """When selecting "AUTO", the second element in the tuple
                   related to the key "x0" should be equal
                   to "RANDOM" or "LHS".""")

        if (not run_dict['x0'][0] == 'AUTO' and not
                run_dict['x0'][0] == 'CUSTOM'):
            raise ValueError(
                """The first element in the tuple related to the key "x0"
                   should be equal to "AUTO" or "CUSTOM".""")

        if not isinstance(run_dict['cx prob'], float):
            raise TypeError(
                """The value of the key "cx prob" should be a positive float,
                   lower or equal to 1.""")
        elif run_dict['cx prob'] > 1.:
            raise ValueError(
                """The value of the key "cx prob" should be a positive float,
                   lower or equal to 1.""")

        if not isinstance(run_dict['mut prob'], float):
            raise TypeError(
                """The value of the key "mut prob" should be a positive float,
                   lower or equal to 1.""")
        elif run_dict['mut prob'] > 1.:
            raise ValueError(
                """The value of the key "mut prob" should be a positive float,
                   lower or equal to 1.""")

        if not isinstance(run_dict['eta'], float):
            raise TypeError(
                """The value of the key "eta" should be a positive float,
                   lower or equal to 1.""")
        elif run_dict['eta'] > 1.:
            raise ValueError(
                """The value of the key "eta" should be a positive float,
                   lower or equal to 1.""")

        if list(run_dict['objectives'].keys())[0] == 'ROB':
            rob = True

    else:

        requirements += ['create only samples',
                         'objective of interest',
                         'draw pdf cdf',
                         ]

        if 'create only samples' not in run_dict:
            run_dict['create only samples'] = False

        if 'draw pdf cdf' not in run_dict:
            run_dict['draw pdf cdf'] = [False]

        for key in requirements[-3:]:
            try:
                run_dict[key]
            except BaseException:
                raise KeyError(
                    '"%s" is missing in the input dictionary.' %
                    key)

        if not isinstance(run_dict['create only samples'], bool):
            raise TypeError(
                """The value of the key "create only samples"
                should be a True or False.""")

        if not isinstance(run_dict['draw pdf cdf'], list):
            raise TypeError(
                'The value of the key "draw pdf cdf" should be a list.')

        if not isinstance(run_dict['draw pdf cdf'][0], bool):
            raise TypeError(
                """The value of the first element in the list "draw pdf cdf"
                should be a True or False.""")

        if run_dict['draw pdf cdf'][0] and not len(
                run_dict['draw pdf cdf']) == 2:
            raise ValueError(
                """When creating the distribution is desired,
                   provide the number of samples in the form of
                   a positive integer as the second list item""")

            if not isinstance(run_dict['draw pdf cdf'][1], int):
                raise TypeError(
                    """When creating the distribution is desired,
                    provide the number of samples in the form of
                    a positive integer as the second list item""")

        if not any(name == run_dict['objective of interest']
                   for name in run_dict['objective names']):
            raise ValueError(
                """The value of the key "objective of interest"
                   should be a string equal to any element
                   in the list with key "objective names".""")

    if rob or uq_bool:

        requirements += ['pol order',
                         'sampling method',
                         'objective names',
                         ]

        if 'sampling method' not in run_dict:
            run_dict['sampling method'] = 'SOBOL'
            warnings.warn("The Sobol' sequence is selected as "
                          "sampling method.")

        for key in requirements[-3:]:
            try:
                run_dict[key]
            except BaseException:
                raise KeyError(
                    '"%s" is missing in the input dictionary.' %
                    key)

        if not isinstance(run_dict['pol order'], int):
            raise TypeError(
                """The value of the key "pol order" should be a
                   positive integer.""")

        if (not run_dict['sampling method'] == 'RANDOM' and not
                run_dict['sampling method'] == 'SOBOL'):
            raise ValueError(
                """The value of the key "sampling method"
                   should be equal to "RANDOM" or "SOBOL".""")

        if not isinstance(run_dict['objective names'], list):
            raise TypeError(
                """The value of the key "objective names"
                   should be a list with the names of the
                   quantities of interest as elements in string format.""")

    if rob:
        requirements += ['objective of interest']

        try:
            run_dict[requirements[-1]]
        except BaseException:
            raise KeyError(
                '"%s" is missing in the input dictionary.' %
                requirements[-1])

        if not isinstance(run_dict['objective of interest'], list):
            raise TypeError(
                """The value of the key "objective of interest"
                   should be a list with the names of the
                   quantities of interest as elements in string format.""")

        if not isinstance(run_dict['objective names'], list):
            raise TypeError(
                """The value of the key "objective names"
                   should be a list with the names of the
                   quantities of interest as elements in string format.""")
        else:
            for name_qoi in run_dict['objective of interest']:
                if not any(name == name_qoi
                           for name in run_dict['objective names']):
                    raise ValueError(
                        """The value of the key "objective of interest"
                           should be a list with strings equal to any element
                           in the list with key "objective names".""")

        if not len(list(run_dict['objectives'].values())[
                   0]) == 2 * len(run_dict['objective of interest']):
            raise ValueError(
                """The number of objectives of interest should be equal to
                           two times the number of weigths. Note that the
                           weigths with uneven indices relate to the standard
                           deviation of the quantity of interest and should
                           therefore be equal to -1 (minimization).""")

    for elem in list(run_dict.keys()):
        if requirements.count(elem) == 0:
            raise ValueError("""" "%s" does not belong in the dictionary.
                                 Try renaming it, such that it corresponds to
                                 an expected dictionary item, or consider
                                 removing it. """ % elem)


def load_case(run_dict, design_space, uq_bool=False, create_only_samples=False):
    """
    For the selected case, the design variables and model parameters
    are loaded based on information from :file:`design_space`.
    In addition, the class object for the selected case is instantiated.

    Parameters
    ----------
    run_dict : dict
        The dictionary with information on the uncertainty quantification.
    design_space : string
        The design_space filename.
    uq_bool : bool, optional
        Indicates if uncertainty quantification is performed.
        The default is False.
    create_only_samples : bool, optional
        Indicates if only samples need to be created. The default is False.

    Returns
    -------
    space_obj : object
        The object with information on the design space and stochastic space
    eval_func : function
        The evaluate function that evaluates the system model.
    params : list, optional
        List with fixed data, used during model evaluation.
    """

    # define case directory
    case_path = os.path.join(os.path.split(os.path.dirname(
                             os.path.abspath(__file__)))[0],
                             'CASES')

    # enlist folders in cases directory
    dir_list = [item for item in os.listdir(case_path) if os.path.isdir(
                os.path.join(
                    case_path,
                    item))]

    # determine the evaluation type
    if uq_bool:
        eval_type = 'UQ'
    elif 'ROB' in run_dict['objectives'].keys():
        eval_type = 'ROB'
    elif 'DET' in run_dict['objectives'].keys():
        eval_type = 'DET'
    else:
        raise ValueError('Error in name of optimization type!')

    # get the StochasticDesignSpace object
    space_obj = StochasticDesignSpace(eval_type,
                                      run_dict['case'],
                                      design_space)

    # in case of optimization, attach objectives
    if not uq_bool:
        space_obj.attach_objectives(run_dict['objectives'])

    if not run_dict['case'] in dir_list:
        raise ValueError('Missing folder: %s folder is not found!' %
                         run_dict['case'])

    if not os.path.isfile(os.path.join(case_path, run_dict['case'],
                                       'case_description.py')):
        raise ValueError('Missing file: case_description.py not found!')

    if not create_only_samples:
        sys.path.insert(0, os.path.join(case_path, run_dict['case']))
        import case_description

        # determine the evaluate function from the considered case directory
        eval_func = case_description.evaluate

        # if fixed parameters for model evaluation are present,
        # get the params list
        if hasattr(case_description, 'set_params'):
            params = case_description.set_params()
        else:
            params = []

    else:
        eval_func = None
        params = None

    return space_obj, eval_func, params


class StochasticDesignSpace(object):
    """
    Class which creates an object which characterizes the stochastic
    design space for the system model evaluation.

    Parameters
    ----------
    opt_type : string
       Defines the type run. Available types = ['DET', 'ROB', 'UQ']
    case : string
        The case name.
    design_space : string
        the design_space file name considered for evaluation

    """

    def __init__(self, opt_type, case, design_space):

        self.case = case
        self.design_space = design_space

        path = os.path.dirname(os.path.abspath(__file__))
        self.case_path = os.path.join(path, case)
        self.opt_type = opt_type
        self.l_b = []
        self.u_b = []
        self.par_dict = {}
        self.var_dict = {}
        self.upar_dict = {}
        self.obj = None

        self.read_design_space()
        self.read_stochastic_space()

    def read_design_space(self):
        """
        Reads the :file:`design_space.csv` file and extracts the deterministic
        values for the parameters and the bounds for the design variables.

        """

        path_to_read = os.path.join(self.case_path, self.design_space)

        # check if design_space file exists
        if not os.path.isfile(path_to_read):
            raise ValueError(
                """Missing file: "design_space.csv" or the name of the case does
                   not exist. Make sure that the name of the case is equal to
                   the name of the folder in CASES.""")

        with open(path_to_read, 'r') as file:
            for line in file:
                tmp = line.split(",")

                # store model parameter names and deterministic value
                if tmp[1] == 'par':
                    if len(tmp) > 3 and tmp[3] != '\n' and tmp[3] != '':
                        raise IndexError(
                            """ Wrong characterization of the parameter %s.
                                Is it supposed to be a variable?
                                Change "par" into "var".""" % tmp[0])

                    self.par_dict[tmp[0]] = float(tmp[2])

                # store design variable names and bounds
                elif tmp[1] == 'var':
                    if len(tmp) < 4:
                        raise IndexError(
                            """ Wrong characterization of the design variable %s.
                                Is it supposed to be a parameter?
                                Change "var" into "par".""" %
                            tmp[0])
                    self.var_dict[tmp[0]] = [float(tmp[2]), float(tmp[3])]
                    self.l_b.append(float(tmp[2]))
                    self.u_b.append(float(tmp[3]))
                    if float(tmp[2]) >= float(tmp[3]):
                        raise NameError(
                            """The lower bound is equal or greater than the
                            upper bound of %s""" % tmp[0])
                else:
                    raise ValueError(
                        """ The line %s does not mention if the
                            characterization corresponds to a parameter ("par")
                            or a design variable ("var").""" % tmp)
        self.n_dim = len(self.var_dict.keys())
        self.n_par = len(self.par_dict.keys())

        # check if the optimization case contains design variables
        if not self.var_dict and 'UQ' not in self.opt_type:
            raise NameError(
                'Variable list is empty! Case cannot be instatiated!')

    def read_stochastic_space(self):
        """
        Reads the :file:`stochastic_space` file and extracts the name,
        distribution and deviation from the mean of the stochastic parameters.

        """

        path_to_read = os.path.join(self.case_path, 'stochastic_space.csv')

        if any(
            x in self.opt_type for x in [
                'ROB',
                'UQ']) and os.path.isfile(path_to_read):

            # get characteristics of stochastic parameters
            with open(path_to_read, 'r') as file:
                for line in file:
                    tmp = line.split(",")
                    self.upar_dict[tmp[0]] = [tmp[1], tmp[2], float(tmp[3])]

        elif (any(x in self.opt_type for x in ['ROB', 'UQ'])
              and not os.path.isfile(path_to_read)):

            raise NameError(
                """Missing file: "stochastic_space.csv".
                   Uncertainty matrix cannot be built!""")

    def attach_objectives(self, obj):
        """
        Attaches the objectives of the configured optimization run.

        Parameters
        ----------
        obj : dict
            the objectives for the optimization run

        """

        self.obj = obj

    def convert_into_dictionary(self, x_i):
        """
        Convert the input sample for model evaluation into a dictionary.

        Parameters
        ----------
        x_i : ndarray
            Input sample for model evaluation.
            x_i = [xp1, xp2, ..., xpm, xd1, xd2, ..., xdn]
            Parameters are included for uncertainty quantification reasons.

        Returns
        -------
        {**parameters, **inputs} : dict
            The dictionary with the variable and parameter names as keys,
            and the input sample values as values.

        """

        # include parameters in first dictionary
        parameters = dict(zip(self.par_dict.keys(), x_i[:len(self.par_dict)]))
        if self.var_dict:
            # when design variables are present, include them in second dict
            inputs = dict(zip(self.var_dict.keys(), x_i[-len(self.var_dict):]))
        else:
            inputs = {}

        # concatenate two dict and return
        return {**parameters, **inputs}



'''
Module to test the uncertainty_quantification module.
'''

import pytest
import rheia.UQ.uncertainty_quantification as uq


@pytest.fixture
def run_dict():
    """
    Generate a generic input dictionary for uncertainty quantification.

    Returns
    -------
    dict_uq : dict
        Input dictionary for uncertainty quantification.

    """

    dict_uq = {'case': 'H2_FUEL',
               'pol order': 2,
               'objective names': ['lcoh', 'mh2'],
               'objective of interest': 'lcoh',
               'sampling method': 'SOBOL',
               'n jobs': 1,
               'draw pdf cdf': [True, 1e5],
               'results dir': 'opt_design_tutorial'
               }

    return dict_uq


def test_get_design_variables(run_dict):
    """
    Assert the design variable dictionary.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for uncertainty quantification.

    """

    var_dict = uq.get_design_variables(run_dict['case'])
    assert var_dict == {'n_pemel': [0., 7.],
                        'n_dcdc_pv': [0., 7.]}


def test_set_design_samples(run_dict):
    """
    Assert the design samples.

    Parameters
    ----------
    run_dict : dict
        Input dictionary for uncertainty quantification.

    """

    var_dict = uq.get_design_variables(run_dict['case'])
    samples = uq.set_design_samples(var_dict, 4)
    assert len(samples) == 4
    assert len(samples[0]) == 2



"""
The :py:mod:`case_description` module contains a function
to read and store the fixed parameters for model evaluation and a function
to evaluate the system model.
"""

import os
import rheia.CASES.H2_POWER.h2_power as pv_h2


def set_params():
    """
    Set the fixed parameters for each model evaluation.

    """

    path = os.path.dirname(os.path.abspath(__file__))

    filename_climate = os.path.join(os.path.abspath(
                                    os.path.join(path,
                                                 os.pardir)),
                                    'DATA',
                                    'climate',
                                    'climate_Brussels.csv')

    filename_demand = os.path.join(os.path.abspath(
                                   os.path.join(path,
                                                os.pardir)),
                                   'DATA',
                                   'demand',
                                   'load_Brussels_dwelling.csv')

    # get the solar irradiance, ambient temperature data and electric demand
    my_data = pv_h2.ReadData(filename_climate, filename_demand)
    sol_irr, t_amb = my_data.load_climate()
    load_elec = my_data.load_demand()

    # store data in params list
    params = [sol_irr, t_amb, load_elec]

    return params


def evaluate(x_in, params=[]):
    '''
    Evaluation of the system objectives for one given design.

    Parameters
    ----------
    x_in : tuple
        An enumerate object for the input sample.
        The first element of x
        - the index of the input sample in the list of samples -
        can be used for multiprocessing purposes of executable files
        with input and output text files.
        The second element of x - the input sample -
        is a dictionary with the names and values for the model parameters
        and design variables.
    params : list, optional
        List with fixed data, used during model evaluation. The default is [].

    Returns
    -------
    lcoe : float
        the levelized cost of electricity
    ssr: float
        the self-sufficiency ratio
    '''

    arguments = params + [x_in[1]]

    # Evaluation object
    my_evaluation = pv_h2.Evaluation(*arguments)

    # evaluate system model
    my_evaluation.evaluation()

    # get values for objectives
    lcoe = my_evaluation.res['lcoe']
    ssr = my_evaluation.res['ssr']

    return lcoe, ssr



"""
The :py:mod:`optimization` module contains functions to load the optimizer,
create the starting samples and run the optimization.
"""

import os
import imp
from shutil import copyfile
from pyDOE import lhs
from rheia.CASES.determine_stoch_des_space import load_case, check_dictionary
import numpy as np
import pandas as pd
#######################
# optimizer functions #
#######################


def parse_available_opt():
    """

    Parse all available optimizers. In this version,
    only NSGA-II is available.

    Returns
    -------
    methods : list of tuples
        The available optimizers with the corresponding classes.

    """

    opt_dir = os.path.join(os.path.split(
        os.path.dirname(
            os.path.abspath(__file__)))[0],
        'OPT')

    # get the optimizer modules
    files = [f for f in os.listdir(opt_dir) if f.endswith('algorithms.py')]
    methods = []
    for file in files:
        obj = imp.load_source(file.split('.')[0], os.path.join(opt_dir, file))
        tmp = obj.return_opt_methods()
        for method in tmp:
            methods.append((method, obj))

    return methods


def load_optimizer(optimizer):
    """

    Load the selected optimizer.

    Parameters
    ----------
    optimizer : string
        The name of the optimizer (currently 'NSGA2').

    Returns
    -------
    opt_obj : object
        The optimization object.

    """

    opt_list = parse_available_opt()
    optimizers = [elem[0] for elem in opt_list]

    # check if optimizer exists and load optimization module
    if optimizer not in optimizers:
        raise KeyError('Optimizer is not available!')

    # get object from optimizer class
    for opt in opt_list:
        if optimizer == opt[0]:
            opt_obj = opt[1].return_opt_obj(optimizer)

    return opt_obj

####################
# starting samples #
####################


def check_existing_results(run_dict, space_obj):
    """
    Check if previously generated results exists in the provided results
    directory.

    Parameters
    ----------
    run_dict : dict
        The dictionary with user-defined values for
        the characterization of the optimization.
    space_obj : object
        The object with information on the design space and stochastic space

    Returns
    -------
    start_from_last_gen : bool
        Boolean that indicates if results exist in the considered result
        directory.

    """

    # path of the population file
    pop_file = os.path.join(os.path.split(os.path.dirname(
        os.path.abspath(__file__)))[0],
        'RESULTS',
        space_obj.case,
        list(run_dict['objectives'].keys())[0], run_dict['results dir'],
        'population.csv')

    # path of the fitness file
    fitness_file = os.path.join(os.path.split(os.path.dirname(
        os.path.abspath(__file__)))[0],
        'RESULTS',
        space_obj.case,
        list(run_dict['objectives'].keys())[0], run_dict['results dir'],
        'fitness.csv')

    # check if the population and fitness file exist
    start_from_last_gen = False
    if os.path.isfile(fitness_file) and os.path.isfile(pop_file):

        start_from_last_gen = True

    return start_from_last_gen


def scale_samples_to_design_space(nondim_doe, space_obj):
    """
    Scales the starting sample to the given design space.

    Parameters
    ----------
    nondim_doe : ndarray
        The non-dimensionalized design of experiment.
    space_obj : object
        The object with information on the design space and stochastic space

    Returns
    -------
    dim_doe : ndarray
        The design of experiment scaled up to the design space.

    """

    dim_doe = np.zeros(nondim_doe.shape)
    for j in range(space_obj.n_dim):
        dim_doe[:, j] = ((space_obj.u_b[j] - space_obj.l_b[j]) *
                         nondim_doe[:, j] + space_obj.l_b[j])

    return dim_doe


def write_starting_samples(doe, filename):
    """

    Writes the starting samples to file.

    Parameters
    ----------
    doe : list
        The set of starting samples.
    filename : string
        The path to the filename for the starting samples.

    """
    with open(filename, 'w') as file:
        for x_in in doe:
            if not isinstance(x_in, list):
                x_in = x_in.tolist()
            for item in x_in:
                file.write('%.8f ' % item)
            file.write('\n')


def create_starting_samples(run_dict, space_obj, start_from_last_gen):
    """

    Load the starting samples for the optimization run.

    Parameters
    ----------
    run_dict : dict
        The dictionary with user-defined values for
        the characterization of the optimization.
    space_obj : object
        The object with information on the design space and stochastic space
    start_from_last_gen : bool
        Boolean that determines if the starting samples
        start from a previously generated population.

    """

    # define folder to put starting samples file
    doe_path = os.path.join(os.path.split(
        os.path.dirname(
            os.path.abspath(__file__)))[0],
        'OPT',
        'INPUTS',
        space_obj.case,
        '%iD' %
        space_obj.n_dim)

    # check if folder exists in INPUTS
    # if not, create one
    if not os.path.exists(doe_path):
        os.makedirs(doe_path)

    # create the doe set of samples
    doe_filename = os.path.join(doe_path,
                                'DOE_n%i' % run_dict['population size'])

    if not start_from_last_gen:
        # if the starting population needs to be created
        if 'AUTO' in run_dict['x0'][0]:

            # generate the starting population randomly
            if 'RANDOM' in run_dict['x0'][1]:
                ddoe = np.random.random(
                    (run_dict['population size'], space_obj.n_dim))

            # generate the starting population
            # based on Latin Hypercube Sampling
            if 'LHS' in run_dict['x0'][1]:
                ddoe = lhs(space_obj.n_dim,
                           samples=run_dict['population size'])

            # scale the starting samples
            doe = scale_samples_to_design_space(ddoe, space_obj)

            # Write starting samples to file
            write_starting_samples(doe, doe_filename)

        else:
            # the starting population is provided in a custom file
            doe_custom = os.path.join(os.path.split(
                os.path.dirname(
                    os.path.abspath(__file__)))[0],
                'CASES',
                space_obj.case,
                run_dict['x0'][1])

            # check if the custom file exists
            if not os.path.isfile(doe_custom):
                raise NameError(
                    """The initial population file %s
                    is not found in the case folder.""" %
                    os.path.basename(doe_custom))

            copyfile(doe_custom, doe_filename)

    else:
        # the starting population is the final generation from the existing
        # population in the result directory
        doe_custom = os.path.join(os.path.split(
            os.path.dirname(
                os.path.abspath(__file__)))[0],
            'RESULTS',
            space_obj.case,
            list(run_dict['objectives'].keys())[0],
            run_dict['results dir'],
            'population.csv')
            
        df = pd.read_csv(doe_custom)
        df.drop(df.tail(1).index,inplace=True)

        rows = df.tail(run_dict['population size']).to_numpy()
        

        doe = []
        for line in rows:
            dummy = []
            for l in line:
                dummy.append(float(l))
            doe.append(dummy)
            
        # check if the size of the existing population matches with the
        # provided one in the optimization dictionary
        if len(doe) == run_dict['population size']:
            write_starting_samples(doe, doe_filename)
        else:
            raise ValueError(""" The defined population size does not match
                                 the population size from a previous run in
                                 this result directory.""")

########################
# run the optimization #
########################


def run_opt(run_dict, design_space='design_space.csv'):
    """
    This function runs the optimization pipeline.
    First, the case, optimizer and configuration
    are loaded. Thereafter, the starting samples
    are created, the specific optimization class
    instantiated and the :py:meth:`run_optimizer`
    method is called.

    Parameters
    ----------
    run_dict : dict
        The dictionary with user-defined values for
        the characterization of the optimization.
    design_space : string, optional
        The design_space filename. The default is 'design_space.csv'.

    """

    # evaluate if the optimization dictionary is properly characterized
    check_dictionary(run_dict)

    # load the object on the design space, the evaluation function
    # and the params provided for each model evaluation
    space_obj, eval_func, params = load_case(run_dict, design_space)

    # add the evaluation function to the optimization dictionary
    run_dict['evaluate'] = eval_func

    # load optimizer class
    opt_class = load_optimizer('NSGA2')

    # check if previous results in this result directory exist
    start_from_last_gen = check_existing_results(run_dict, space_obj)

    # define starting samples
    create_starting_samples(run_dict, space_obj, start_from_last_gen)

    # optimizer object
    res = opt_class(run_dict, space_obj, start_from_last_gen, params)

    # run the optimizer
    res.run_optimizer()



"""
The :py:mod:`post_process` module contains classes to present the
optimization and uncertainty quantification results.
"""

import os
import numpy as np


class PostProcessOpt():
    """
    The PostProcessOpt class provides methods to retrieve the fitness and
    population for the generations provided during the optimization.

    Parameters
    ----------
    case : str
        Name of the case to be evaluated.
    eval_type : str
        The type of evaluation performed. Equals 'DET' for deterministic
        optimization and 'ROB' for robust optimization.

    """

    def __init__(self, case, eval_type):
        path_file = os.path.dirname(os.path.abspath(__file__))
        path_start = os.path.abspath(os.path.join(path_file, os.pardir))
        path = os.path.join(path_start,
                                 'RESULTS',
                                 case
                                 )
        self.n_pop = 0
        self.x_lines = []
        self.y_lines = []
        self.result_path = os.path.join(path,
                                        eval_type,
                                        )
        self.fitness_file = ''
        self.population_file = ''
        self.n_gen = 0

    def determine_pop_gen(self):
        """

        Determines the number of design samples in the population
        and the number of generations performed.

        """

        self.n_gen = 0
        with open(self.fitness_file, 'r') as file:
            self.y_lines = file.readlines()
            for string in self.y_lines:
                if string == '-\n':
                    # count the number of generations
                    self.n_gen += 1

            # determine the population size
            self.n_pop = int(len(self.y_lines) / self.n_gen - 1)

    def get_fitness_values(self, gen):
        """

        Returns the fitness values for the population
        generated in the specified generation.

        Parameters
        ----------
        gen : int
            The generation of interest.

        Returns
        -------
        fit_val : ndarray
            The fitness values for the population of interest.

        """

        fit_val = np.zeros((len(self.y_lines[0].split(",")), self.n_pop))
        for index, line in enumerate(
                self.y_lines[(gen - 1) * self.n_pop + gen -1:
                             gen * self.n_pop-1 + gen]):

            # add fitness values for the desired generation
            fit_val[:, index] = [float(i) for i in line.split(",")]

        return fit_val

    def get_population_values(self, gen):
        """
        Returns the fitness values for the population
        generated in the specified generation.

        Parameters
        ----------
        gen : int
            The generation of interest.

        Returns
        -------
        pop_val : ndarray
            The population of interest.

        """

        with open(self.population_file, 'r') as file:
            self.x_lines = file.readlines()

        pop_val = np.zeros((len(self.x_lines[0].split(",")), self.n_pop))
        for index, line in enumerate(
                self.x_lines[(gen - 1) * self.n_pop + gen-1:
                             gen * self.n_pop-1 + gen]):

            # add population values for the desired generation
            pop_val[:, index] = [float(i) for i in line.split(",")]

        return pop_val

    def sorted_result_file(self, y_val, x_val):
        """

        Generates the files that include the sorted population
        and fitness files. The population and fitness are sorted
        based on the first objective.

        Parameters
        ----------
        y_val : ndarray
            The fitness values.
        x_val : ndarray
            The set of design samples.

        """

        # write sorted fitness values in corresponding file
        with open(self.fitness_file[:-4] + "_final_sorted.csv", 'w') as file:
            for sample in y_val:
                for value in sample:
                    file.write('%f,' % value)
                file.write('\n')

        # write sorted population values in corresponding file
        with open(self.population_file[:-4] + "_final_sorted.csv", 'w') as file:
            for sample in x_val:
                for value in sample:
                    file.write('%f, ' % value)
                file.write('\n')

    def get_fitness_population(self, result_dir, gen=0):
        """

        Returns the population and corresponding fitness values
        for the generation of interest.

        Parameters
        ----------
        result_dir : str
            The directory were the results are stored.
        gen : int, optional
            The generation of interest. The default is 0,
            i.e. the final generation.

        Returns
        -------
        y_gen : ndarray
            The fitness values.
        x_gen : ndarray
            The set of design samples.

        """

        self.fitness_file = os.path.join(self.result_path,
                                         result_dir,
                                         'fitness.csv',
                                         )

        self.population_file = os.path.join(self.result_path,
                                            result_dir,
                                            'population.csv',
                                            )

        # set the population size and number of generations
        self.determine_pop_gen()

        # get fitness values for desired generation
        y_unsorted = self.get_fitness_values(gen)

        # get population values for desired generation
        x_unsorted = self.get_population_values(gen)

        a, b = x_unsorted.shape
        c, d = y_unsorted.shape

        # get indices when sorting the y_unsorted array,
        # sorting is based on first objective
        indices = np.argsort(y_unsorted[0])

        x_gen = np.zeros((a, b))
        y_gen = np.zeros((c, d))
        for j, k in enumerate(indices):
            for index, y_in in enumerate(y_gen):
                y_gen[index][j] = y_unsorted[index][k]
            for index, x_in in enumerate(x_gen):
                x_gen[index][j] = x_unsorted[index][k]

        # print sorted results in separate files
        self.sorted_result_file(y_gen.transpose(), x_gen.transpose())

        return y_gen, x_gen


class PostProcessUQ():
    """
    The PostProcessUQ class provides methods to retrieve the LOO error,
    plot the Sobol indices, PDF and CDF.

    Parameters
    ----------
    case : str
        Name of the case to be evaluated.
    pol_order : int
        The polynomial order.

    """

    def __init__(self, case, pol_order):
        path_file = os.path.dirname(os.path.abspath(__file__))
        path_start = os.path.abspath(os.path.join(path_file, os.pardir))
        path = os.path.join(path_start,
                                 'RESULTS',
                                 case
                                 )
        self.result_path = os.path.join(path,
                                        'UQ',
                                        )

        self.pol_order = pol_order

    def read_distr_file(self, distr_file):
        """

        Reads the file with information on the
        cumulative density function or probability
        density function.

        Parameters
        ----------
        distr_file : str
            The name of the distribution file.

        Returns
        -------
        x_val : ndarray
            The values from the PDF or CDF on the
            quantity of interest.
        y_val : ndarray
            The probability density (for the PDF)
            or cumulative probability (for the CDF).

        """

        # read the file with info on the CDF or PDF
        with open(distr_file, 'r') as file:
            lines = file.readlines()
            x_val = np.ones(len(lines) - 1)
            y_val = np.ones(len(lines) - 1)
            for index, line in enumerate(lines[1:]):
                tmp = line.split(",")
                x_val[index] = float(tmp[0])
                y_val[index] = float(tmp[1])

        return x_val, y_val

    def get_sobol(self, result_dir, objective):
        """

        Retrieves the information on the Sobol' indices from
        the corresponding file in the result directory.

        Parameters
        ----------
        result_dir : str
            The result directory.
        objective : str
            The name of the quantity of interest.

        Returns
        -------
        names : list
            The names of the stochastic parameters.
        sobol : list
            The total' order Sobol' indices.

        """

        sobol_file = os.path.join(self.result_path,
                                  '%s' % result_dir,
                                  'full_pce_order_%i_%s_Sobol_indices.csv' % (
                                      self.pol_order, objective)
                                  )

        # retrieve the parameter names and corresponding Sobol indices
        res_tmp = []
        with open(sobol_file, 'r') as file:
            for line in file.readlines()[1:]:
                res_tmp.append([i for i in line.split(",")])
            names = [row[0] for row in res_tmp]
            sobol = [float(row[2]) for row in res_tmp]

        return names, sobol

    def get_pdf(self, result_dir, objective):
        """

        Retrieves the points that define the probability density function.

        Parameters
        ----------
        result_dir : str
            The result directory.
        objective : str
            The name of the quantity of interest.

        Returns
        -------
        x_val : ndarray
            The values from the PDF on the
            quantity of interest.
        y_val : ndarray
            The probability density.

        """

        pdf_file = os.path.join(self.result_path,
                                '%s' % result_dir,
                                'data_pdf_%s.csv' % objective
                                )

        # get the x and y values for the PDF
        x_val, y_val = self.read_distr_file(pdf_file)

        return x_val, y_val

    def get_cdf(self, result_dir, objective):
        """

        Retrieves the points that define the cumulative density function.

        Parameters
        ----------
        result_dir : str
            The result directory.
        objective : str
            The name of the quantity of interest.

        Returns
        -------
        x : ndarray
            The values from the CDF on the
            quantity of interest.
        y : ndarray
            The cumulative probability.

        """

        cdf_file = os.path.join(self.result_path,
                                '%s' % result_dir,
                                'data_cdf_%s.csv' % objective
                                )

        # get the x and y values for the CDF
        x_val, y_val = self.read_distr_file(cdf_file)

        return x_val, y_val

    def get_loo(self, result_dir, objective):
        """

        Reads the Leave-One-Out error from the corresponding
        file in the result directory.

        Parameters
        ----------
        result_dir : str
            The result directory.
        objective : str
            The name of the quantity of interest.

        Returns
        -------
        loo : float
            The Leave-One-Out error.

        """

        loo_file = os.path.join(self.result_path,
                                '%s' % (result_dir),
                                'full_pce_order_%i_%s.txt' % (
                                    self.pol_order, objective)
                                )

        # retrieve the LOO error
        with open(loo_file, 'r') as file:
            line = file.readlines()[0]
            loo = float(line.split()[1])

        return loo

    def get_mean_std(self, result_dir, objective):
        """

        Reads the mean and standard deviation
        from the corresponding file in the result directory.

        Parameters
        ----------
        result_dir : str
            The result directory.
        objective : str
            The name of the quantity of interest.

        Returns
        -------
        mean : float
            The mean.
        std : float
            The standard deviation.

        """

        loo_file = os.path.join(self.result_path,
                                '%s' % (result_dir),
                                'full_pce_order_%i_%s.txt' % (
                                    self.pol_order, objective)
                                )

        # retrieve the mean and standard deviation
        with open(loo_file, 'r') as file:
            line = file.readlines()[2]
            mean = float(line.split()[1])

        with open(loo_file, 'r') as file:
            line = file.readlines()[3]
            std = float(line.split()[2])

        return mean, std



    def get_max_sobol(self, result_dirs, objective, threshold=0.05):
        """

        This method gathers the Sobol' indices for each stochastic parameter
        for each sample. The highest Sobol' index for each stochastic
        parameter is compared with the threshold value. If the highest
        Sobol' index is higher than the threshold, the name of the
        stochastic parameter is printed under 'significant Sobol indices'.
        If not, it is printed under 'negligible Sobol indices'.

        Parameters
        ----------
        result_dir : list
            The result directories.
        objective : str
            The name of the quantity of interest.
        threshold : float, optional
            The threshold that determines if a Sobol' index
            is considered significant. The default is 0.05.

        """

        # store the dictionary with parameter names and Sobol indices in a
        # list for each result directory evaluated
        n_samples = len(result_dirs)
        res_dict = [{}] * n_samples
        for index, result_dir in enumerate(result_dirs):
            names, sobol = self.get_sobol(result_dir, objective)
            res_dict[index] = dict(zip(names, sobol))

        # get the highest Sobol index for each parameter over the different
        # dictionaries
        max_dict = dict()
        for name in names:
            sobol_res = np.zeros(n_samples)
            for j, dic in enumerate(res_dict):
                sobol_res[j] = dic[name]
            max_dict[name] = max(sobol_res)

        print('significant Sobol indices:')
        for k in names:
            if max_dict[k] >= threshold:
                print('%s: %4f' % (k, max_dict[k]))

        print('\nnegligible Sobol indices:')
        for k in names:
            if max_dict[k] < threshold:
                print('%s: %4f' % (k, max_dict[k]))



"""
The :py:mod:`four_bar_truss` module contains a function
to evaluate the four bar truss system model.
"""

def four_bar_truss(x_i):
    """

    This function evaluates the volume and displacement
    of a four-bar truss system.

    Parameters
    ----------
    x_i : array
        The input sample.

    Returns
    -------
    vol : float
        The volume of the truss.
    disp : float
        The displacement of the node.

    """

    # the volume
    vol = x_i['L'] * (2. * x_i['A_1'] + 2.**(0.5) *
                       x_i['A_2'] + x_i['A_3']**(0.5) + x_i['A_4'])

    # the displacement of the outer node
    disp = x_i['F'] * x_i['L'] * (2. / (x_i['A_1'] * x_i['E_1']) +
                                  2. * 2**(0.5) / (x_i['A_2'] * x_i['E_2']) -
                                  2. * 2**(0.5) / (x_i['A_3'] * x_i['E_3']) +
                                  2. / (x_i['A_4'] * x_i['E_4']))

    return vol, disp



"""
The :mod:`uncertainty_quantification` module provides functions to
execute uncertainty quantification.

"""

import os
from pyDOE import lhs
from rheia.CASES.determine_stoch_des_space import load_case, check_dictionary
import rheia.UQ.pce as uq


def get_design_variables(case):
    """
    This function loads the design variable names and bounds
    out of the :file:`design_space.csv` file.

    Parameters
    ----------
    case : string
        The name of the case.


    Returns
    -------
    var_dict : dict
        A dictionary which includes the design variables and their bounds.

    """
    var_dict = {}

    path = os.path.dirname(os.path.abspath(__file__))
    path_to_read = os.path.join(
        os.path.abspath(
            os.path.join(
                path,
                os.pardir)),
        'CASES',
        case,
        'design_space.csv')

    # read in the design variable bounds
    with open(path_to_read, 'r') as file:
        for line in file:
            tmp = line.split(",")
            if tmp[1] == 'var':
                var_dict[tmp[0]] = [float(tmp[2]), float(tmp[3])]

    return var_dict


def set_design_samples(var_dict, n_samples):
    """
    Based on the design variable characteristics,
    a set of design samples is created through
    Latin Hypercube Sampling.

    Parameters
    ----------
    var_dict : dict
        A dictionary which includes the design variables and their bounds.
    n_samples : int
        The number of design samples to be created.

    Returns
    -------
    samples : array
        The generated design samples.

    """

    # generate the samples through Latin Hypercube Sampling
    samples = lhs(len(var_dict), samples=n_samples)

    bounds = list(var_dict.values())

    # scale the samples based on the design variable bounds
    for i, bound in enumerate(bounds):
        samples[:, i] *= (bound[1] - bound[0])
        samples[:, i] += bound[0]

    return samples


def write_design_space(case, iteration, var_dict, sample, ds = 'design_space.csv'):
    """
    A new design space file is created. In this file,
    the model parameters are copied from the original file,
    i.e. file:`design_space`. The design variable names are copied,
    but the bounds are loaded out of the array `sample`.
    This function is of interest when evaluating the LOO error
    or Sobol' indices for several design samples.

    Parameters
    ----------
    case : string
        The name of the case.
    iteration : int
        The index of the design sample
        out of the collection of generated design samples.
    var_dict : dict
        A dictionary which includes the design variables and their bounds.
    sample : array
        The design sample out of the collection of generated design samples.
    ds : string, optional
        The design_space filename. The default is 'design_space'.

    """

    path = os.path.dirname(os.path.abspath(__file__))

    des_var_file = os.path.join(os.path.abspath(os.path.join(path, os.pardir)),
                                'CASES',
                                case,
                                'design_space.csv',
                                )

    new_des_var_file = os.path.join(
        os.path.abspath(
            os.path.join(
                path,
                os.pardir)),
        'CASES',
        case,
        '%s_%i%s' % (ds[:-4], iteration, ds[-4:])
    )

    # write the new design_space file if it does not exist already
    if not os.path.isfile(new_des_var_file):
        with open(des_var_file, 'r') as file:
            text = []
            for line in file.readlines():
                found = False
                tmp = line.split()
                for index, name in enumerate(list(var_dict.keys())):

                    if name == tmp[0]:
                        text.append('%s par %f \n' % (name, sample[index]))
                        found = True
                if not found:
                    text.append(line)

        with open(new_des_var_file, 'w') as file:
            for item in text:
                file.write("%s" % item)


def run_uq(run_dict, design_space='design_space.csv'):
    """
    This function is the main to run uncertainty quantification.
    First, the input distributions are created,
    followed by the reading of previously evaluated samples.
    Thereafter, the new samples are created and evaluated
    in the system model when desired. Finally, the PCE is
    constructed, the statistical moments printed and
    the distributions generated (when desired) for the
    quantity of interest.

    Parameters
    ----------
    run_dict : dict
        The dictionary with information on the uncertainty quantification.
    design_space : string, optional
        The design_space filename. The default is 'design_space'.

    """

    # check if the UQ dictionary is properly characterized
    check_dictionary(run_dict, uq_bool=True)

    objective_position = run_dict['objective names'].index(
        run_dict['objective of interest'])

    # load the object on the design space, the evaluation function
    # and the params provided for each model evaluation
    space_obj, eval_func, params = load_case(
        run_dict, design_space, uq_bool=True,
        create_only_samples=run_dict['create only samples'])

    my_data = uq.Data(run_dict, space_obj)

    # acquire information on stochastic parameters
    my_data.read_stoch_parameters()

    # create result csv file to capture all input-output of the samples
    my_data.create_samples_file()

    # create experiment object
    my_experiment = uq.RandomExperiment(my_data, objective_position)

    # create uniform/gaussian distributions and corresponding orthogonal
    # polynomials
    my_experiment.create_distributions()

    # calculate the number of terms in the PCE according to the
    # truncation scheme
    my_experiment.n_terms()

    # read in the previously generated samples
    my_experiment.read_previous_samples(run_dict['create only samples'])

    # create a design of experiment for the remaining samples
    # to be evaluated
    my_experiment.create_samples(
        size=my_experiment.n_samples - len(my_experiment.x_prev))

    # check if the samples need to be evaluated or not
    my_experiment.create_only_samples(run_dict['create only samples'])

    # when the PCE needs to be constructed
    if not run_dict['create only samples']:

        # evaluate the samples remaining to reach the required
        # number of samples for the PCE
        if my_experiment.n_samples > len(my_experiment.x_prev):
            my_experiment.evaluate(eval_func, params)
        elif my_experiment.n_samples == len(my_experiment.x_prev):
            my_experiment.y = my_experiment.y_prev
        else:
            my_experiment.y = my_experiment.y_prev[:my_experiment.n_samples]

        # create PCE object
        my_pce = uq.PCE(my_experiment)

        # evaluate the PCE
        my_pce.run()

        # calculate the LOO error
        my_pce.calc_loo()

        # calculate the Sobol' indices
        my_pce.calc_sobol()

        # extract and print results
        my_pce.print_res()

        # generate the pdf and cdf when desired
        if run_dict['draw pdf cdf'][0]:
            my_pce.draw(int(run_dict['draw pdf cdf'][1]))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(name='rheia',
      version='1.1.0',
      description='Robust design optimization of renewable Hydrogen and dErIved energy cArrier systems',
      url='https://github.com/rheia-framework/RHEIA',
      author='Diederik Coppitters, Panagiotis Tsirikoglou, Ward De Paepe, Konstantinos Kyprianidis, Anestis Kalfas, Francesco Contino',
      author_email='rheia.framework@gmail.com',
      package_dir={"": "src"},
      packages= setuptools.find_packages(where="src"),
      classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
      ],       
      install_requires=[
      'pyDOE',
      'deap',
      'numpy',
      'scipy',
      'sobolsequence'
      ],
      python_requires = ">=3.6",
      include_package_data=True,
      zip_safe=False)


"""
The :py:mod:`case_description` module contains a function
to evaluate the four bar truss system model.
"""

from rheia.CASES.FOUR_BAR_TRUSS.four_bar_truss import four_bar_truss

def evaluate(x_in, params=[]):
    '''
    Evaluation of the system objectives for one given design.

    Parameters
    ----------
    x_in : tuple
        An enumerate object for the input sample.
        The first element of x
        - the index of the input sample in the list of samples -
        can be used for multiprocessing purposes of executable files
        with input and output text files.
        The second element of x - the input sample -
        is a dictionary with the names and values for the model parameters
        and design variables.
    params : list, optional
        List with fixed data, used during model evaluation. The default is [].

    Returns
    -------
    vol : float
        truss volume
    disp : float
        displacement of outer node
    '''

    # evaluate four bar truss model
    vol, disp = four_bar_truss(x_in[1])

    return vol, disp



"""
The :py:mod:`run_energyplan` module is a wrapper function to run
the EnergyPLAN model.
"""

import subprocess as sp
import os


def create_new_input_file(input_file, new_input_file, x):
    """

    This function writes a new input file for the EnergyPLAN model.
    The new input file copies the lines from the original input file,
    but replaces the default values related to the model parameters
    with the new values generated by the UQ or optimization algorithm.

    Parameters
    ----------
    input_file : str
        The path of the default input file.
    new_input_file : str
        The path of the new input file.
    x : dict
        The new values for the model parameters.

    """

    with open(input_file, 'r') as f:
        lines = f.readlines()

    skip_line = False
    with open(new_input_file, 'w') as f:
        for line in lines:
            if not skip_line:
                if str(line) == 'Input_el_demand_Twh=\n':
                    f.write(line)
                    f.write(str(x['elec_demand']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_hp2_cop=\n':
                    f.write(line)
                    f.write(str(x['COP']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_ELTtrans_fuel=\n':
                    f.write(line)
                    f.write(str(x['eff_H2']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_chp2_el=\n':
                    f.write(line)
                    f.write(str(x['CHP_eff_el']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_chp3_el=\n':
                    f.write(line)
                    f.write(str(x['CHP_eff_el']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_chp2_th=\n':
                    f.write(line)
                    f.write(str(x['CHP_eff_th']) + '\n')
                    skip_line = True
                elif str(line) == 'input_eff_chp3_th=\n':
                    f.write(line)
                    f.write(str(x['CHP_eff_th']) + '\n')
                    skip_line = True
                else:
                    f.write(line)
            else:
                skip_line = False


def read_output_file(result_file):
    """

    This function reads the output file, generated by
    the EnergyPLAN model. From this output file, the
    values for the CO2-emission and fuel consumption
    are stored and returned.

    Parameters
    ----------
    result_file : str
        Path of the result file.

    Returns
    -------
    co2 : float
        The CO2-emission returned by the EnergyPLAN model.
    fuel : float
        The fuel consumption returned by the EnergyPLAN model.

    """
    with open(result_file, 'r') as f:
        lines = f.readlines()

    co2_str = ''
    fuel_str = ''
    for line in lines:
        if 'CO2-emission (total)' in line:
            for j in line[40:46]:
                if j == ',':
                    j = '.'
                co2_str += j
                co2 = float(co2_str)
        if 'Fuel Consumption (total)' in line:
            for j in line[40:46]:
                if j == ',':
                    j = '.'
                fuel_str += j
                fuel = float(fuel_str)

    return co2, fuel


def energyplan(x_tup):
    """

    This function generates a new input file, evaluates
    this input file in the EnergyPLAN model and reads
    the values for the quantities of interest from the
    generated output file. These values are returned.

    Parameters
    ----------
    x_tup : tuple
        The input sample with index that corresponds to
        the position of the sample in the list of samples.

    Returns
    -------
    co2 : float
        The CO2-emission returned by the EnergyPLAN model.
    fuel : float
        The fuel consumption returned by the EnergyPLAN model.

    """

    x = x_tup[1]
    index = x_tup[0]
    path = os.path.dirname(os.path.abspath(__file__))

    ep_path = r'C:\energyPLAN\energyPLAN.exe'
    input_file = os.path.join(path, 'case.txt')

    new_input_file = '%s_%i.txt' % (input_file[:-4], index)

    # create a new input file for evaluation in EnergyPLAN
    create_new_input_file(input_file, new_input_file, x)

    result_file = os.path.join(path, 'result_%i.txt' % index)

    # evaluate the EnergyPLAN model
    cm = [ep_path, '-i', new_input_file, '-ascii', result_file]
    sp.call(cm)

    # read the output file, generated by the EnergyPLAN model
    co2, fuel = read_output_file(result_file)

    os.remove(result_file)
    os.remove(new_input_file)

    return co2, fuel
